﻿using System;
using System.CodeDom;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.Bandits {
  // helper to create canonical forms of expressions
  // TODO: change symbolicregressionpoly10problem to use this class
  // this does not support expressions with constants (in transformations we assume constant opt is used)
  // (e.g. we transform all negative signs to + because it is assumed a negative constant can be produced for the term)
  public class ExpressionExtender {

    // supports the grammar
    // G(E):
    //     E -> V | V+E | V-E | V*E | V%E | (E) 
    //     V -> <variables>
    //     ";

    // transformations: 
    // - transform c*b*a to a*b*c (lexicographic ordering for factors, cached)
    // - transfrom - to + (assumed optimized constants for terms)

    private string sentence;
    private int syIdx;

    public string CanonicalRepresentation(string phrase) {
      InitLex(phrase);
      var e = CanonicalExpr();
      return e.ToString();
    }


    private void InitLex(string sentence) {
      this.sentence = sentence;
      this.syIdx = 0;
    }

    private char CurSy() {
      if (syIdx >= sentence.Length) return '\0';
      return sentence[syIdx];
    }
    private void NewSy() {
      if (syIdx < sentence.Length) syIdx++;
    }

    // an expression is a list of terms 
    // a term is an (ordered) list of factors
    // a factor is a symbol or an inverted expression
    // CanonicalExpression reads multiple terms (expressions) and must merge the terms in the expression (ordering and duplicates)
    // CanonicalTerm reads multiple factors and must expand factors whenever it reads a combined factor (expression) it produces an expression
    // CanonicalFactor produces an expression 


    // canonical expression returns either a single term or a set of terms
    private Expr CanonicalExpr() {
      var terms = new List<Expr>();
      terms.Add(CanonicalTerm());
      var curSy = CurSy();
      while (curSy == '+' || curSy == '-' || curSy == '^') {
        if (curSy == '+') {
          NewSy();
          terms.Add(CanonicalTerm());
        } else if (curSy == '-') {
          NewSy();
          terms.Add(CanonicalTerm()); // minus is the same as plus assuming constant opt
        } else {
          NewSy();
          throw new NotImplementedException();
          // var e = Expr(variables, constants);
          // r = Not(r) * e + r * Not(e); // xor = (!x AND y) OR (x AND !y)
        }
        curSy = CurSy();
      }

      return new Expr(terms.SelectMany(t => t.Terms));
    }

    // canonical term returns either a single term (product of factors) or a set of terms 
    private Expr CanonicalTerm() {
      var factors = new List<Expr>();
      var invFactors = new List<Expr>();
      var f = CanonicalFact();
      if (f != null) factors.Add(f);
      var curSy = CurSy();
      while (curSy == '*' || curSy == '%') {
        if (curSy == '*') {
          NewSy();
          f = CanonicalFact();
          if (f != null) factors.Add(f);
        } else {
          NewSy();
          f = CanonicalFact();
          f.Invert();
          if (f != null) invFactors.Add(f);
        }
        curSy = CurSy();
      }

      // cancellation
      foreach (var invF in invFactors.ToArray()) {
        invF.Invert();
        if (factors.Contains(invF)) {
          invFactors.Remove(invF);
          factors.Remove(invF);
          if (factors.Count == 0) factors.Add(new Expr(new Term('1')));
        } else invF.Invert();
      }

      return ExpandFactors(factors, invFactors);
    }

    private Expr ExpandFactors(List<Expr> factors, List<Expr> invFactors) {
      // if (invFactors.Count > 0) throw new NotImplementedException();
      Expr currentFact = factors.First(); // new Expr(simpleFactor));
      Debug.Assert(!currentFact.Inverse); // the first factor is never an inverted factor
      foreach (var fact in factors.Skip(1)) {
        currentFact = AllProducts(currentFact, fact);
      }
      foreach (var invF in invFactors) {
        currentFact = AllInvProducts(currentFact, invF);
      }
      return currentFact;
    }

    private Expr AllProducts(Expr a, Expr b) {
      var aTerms = a.Terms.ToArray();
      var bTerms = b.Terms.ToArray();
      var combs = from aT in aTerms
                  from bT in bTerms
                  select new Term(aT.Symbols.Concat(bT.Symbols), aT.InvExpressions.Concat(bT.InvExpressions));
      return new Expr(combs);
    }

    private Expr AllInvProducts(Expr a, Expr b) {
      var aTerms = a.Terms.ToArray();
      var combs = from aT in aTerms
                  select new Term(aT.Symbols, aT.InvExpressions.Concat(new Expr[] { b }));
      return new Expr(combs);
    }

    // canonical fact returns a factor (either a singe variable, or a set of terms)
    private Expr CanonicalFact() {
      var curSy = CurSy();
      if (curSy == '!') {
        throw new NotSupportedException();
      } else if (curSy == '(') {
        NewSy();
        Expr r = CanonicalExpr();
        if (CurSy() != ')') throw new ArgumentException();
        NewSy();
        return r;
      } else if (curSy >= 'a' && curSy <= 'z') {
        NewSy();
        return new Expr(new Term(curSy));
        //       } else if (curSy >= '0' && curSy <= '9') {
      } else if (curSy >= 'A' && curSy <= 'Z') {
        // treat nonterminals in the same way as variables
        NewSy();
        return new Expr(new Term(curSy));

      } else throw new ArgumentException("found symbol " + curSy);
    }

    #region term
    // term can be merged (essentially an ordered list of factors)
    internal class Term : IComparable<Term> {
      private readonly SortedList<char, int> factorSymbs; // factor symbol and the number of occurrences
      private readonly SortedList<Expr, int> invExpressions;

      public IEnumerable<char> Symbols {
        get {
          return factorSymbs.SelectMany(p => Enumerable.Repeat(p.Key, p.Value));
        }
      }

      public IEnumerable<Expr> InvExpressions {
        get {
          if (invExpressions == null) return Enumerable.Empty<Expr>();
          return invExpressions.Where(p => p.Value > 0).SelectMany(p => Enumerable.Repeat(p.Key, p.Value));
        }
      }

      public Term(char f) {
        factorSymbs = new SortedList<char, int>(new FactorComparer());
        factorSymbs.Add(f, 1);
      }
      public Term(IEnumerable<char> factors, IEnumerable<Expr> invFactors) {
        factorSymbs = new SortedList<char, int>(new FactorComparer());
        invExpressions = new SortedList<Expr, int>();
        foreach (var f in factors) {
          if (factorSymbs.ContainsKey(f)) factorSymbs[f] += 1;
          else factorSymbs.Add(f, 1);
        }

        foreach (var invF in invFactors) {
          if (invF.Terms.Count == 1) {
            var t = invF.Terms.First();
            foreach (var f in factors.Concat(new char[] { '1' })) {
              if (t.factorSymbs.ContainsKey(f)) {
                t.factorSymbs[f] -= 1;
                if (t.factorSymbs[f] == 0) t.factorSymbs.Remove(f);
              }
            }
          }
          if (invExpressions.ContainsKey(invF)) invExpressions[invF] += 1;
          else invExpressions.Add(invF, 1);
        }
      }

      public int CompareTo(Term other) {
        if (ContainsNonTerminal(Symbols) && !ContainsNonTerminal(other.Symbols)) {
          return 1;
        } else if (!ContainsNonTerminal(Symbols) && ContainsNonTerminal(other.Symbols)) {
          return -1;
        } else {
          var countComp = Symbols.Count().CompareTo(other.Symbols.Count());
          if (countComp != 0) return countComp;
          return string.Join("", Symbols).CompareTo(string.Join("", other.Symbols));
        }
      }

      public override string ToString() {
        if (InvExpressions.Any())
          return string.Join("*", Symbols) + "%" + string.Join("%", InvExpressions);
        else {
          return string.Join("*", Symbols);
        }
      }
      public override bool Equals(object obj) {
        var other = obj as Term;
        if (other == null) return false;
        if (this.Symbols.Count() != other.Symbols.Count()) return false;
        if (this.InvExpressions.Count() != other.InvExpressions.Count()) return false;
        if (this.Symbols.Zip(other.Symbols, Tuple.Create).Any(t => t.Item1 != t.Item2)) return false;
        if (this.InvExpressions.Zip(other.InvExpressions, Tuple.Create).Any(t => t.Item1 != t.Item2)) return false;
        return true;
      }
      public override int GetHashCode() {
        var h = 31415;
        foreach (var v in Symbols) {
          h ^= v.GetHashCode();
        }
        foreach (var v in InvExpressions) {
          h ^= v.GetHashCode();
        }
        return h;
      }
    }
    #endregion

    #region expr

    internal class Expr : IComparable<Expr> {
      public readonly SortedSet<Term> Terms; // only set for Kind == Expr 
      public bool Inverse;
      public Expr(Term t) {
        Terms = new SortedSet<Term>();
        Terms.Add(t);
      }
      public Expr(IEnumerable<Term> exprTerms) {
        Terms = new SortedSet<Term>();
        foreach (var t in exprTerms) {
          Terms.Add(t);
        }
      }

      public void Merge(Expr other) {
        this.Terms.UnionWith(other.Terms);
      }

      public int CompareTo(Expr other) {
        var sizeComp = this.Terms.Count.CompareTo(other.Terms.Count);
        if (sizeComp != 0) return sizeComp;
        // same size => compare terms
        foreach (var pair in Terms.Zip(other.Terms, Tuple.Create)) {
          var termComp = pair.Item1.CompareTo(pair.Item2);
          if (termComp != 0) return termComp;
        }
        return 0;
      }

      public override string ToString() {
        if (Inverse && Terms.Count > 1)
          return "(" + string.Join("+", Terms) + ")";
        else if (Inverse && Terms.Count == 1) {
          return Terms.First().ToString().Replace("%", "#").Replace("*", "%").Replace("#", "*");
        } else
          return string.Join("+", Terms);
      }
      public override bool Equals(object obj) {
        var other = obj as Expr;
        if (other == null) return false;
        if (this.Inverse != other.Inverse) return false;
        if (this.Terms.Count() != other.Terms.Count()) return false;
        return this.Terms.Intersect(other.Terms).Count() == this.Terms.Count;
      }
      public override int GetHashCode() {
        var h = 31415;
        if (Terms != null)
          foreach (var t in Terms) {
            h ^= t.GetHashCode();
          }
        return h;
      }

      public void Invert() {
        this.Inverse = !Inverse;
      }
    }
    #endregion
    internal static bool IsNonTerminal(char symb) {
      return symb >= 'A' && symb <= 'Z';
    }
    internal static bool ContainsNonTerminal(IEnumerable<char> symbs) {
      return symbs.Any(IsNonTerminal);
    }

    internal class FactorComparer : IComparer<char> {
      public int Compare(char x, char y) {
        if (IsNonTerminal(x) && !IsNonTerminal(y)) {
          return 1;
        } else if (!IsNonTerminal(x) && IsNonTerminal(y)) {
          return -1;
        } else if (IsNonTerminal(x) && IsNonTerminal(y)) {
          return x.CompareTo(y);
        } else {
          return x.CompareTo(y);
        }
      }
    }
  }
}
